/*
 * File:        computeDistanceFunction2d.c
 * Copyright:   (c) 2005-2006 Kevin T. Chu
 * Revision:    $Revision: 1.12 $
 * Modified:    $Date: 2006/09/18 16:17:01 $
 * Description: MATLAB MEX-file for using the fast marching method to
 *              compute the distance function for 2d level set functions
 */

/*===========================================================================
 *
 * computeDistanceFunction2d() computes a distance from an 
 * arbitrary level set function in two-dimensions using the 
 * Fast Marching Method.
 * 
 * Usage: distance_fcn = computeDistanceFunction2d(phi, dX, ...
 *                                                 mask, ...
 *                                                 spatial_derivative_order)
 *
 * Arguments:
 * - phi:                       level set function to use in 
 *                                computing distance function
 * - dX:                        array containing the grid spacing
 *                                in each coordinate direction
 * - mask:                      mask for domain of problem;
 *                                grid points outside of the domain
 *                                of the problem should be set to a
 *                                negative value
 *                                (default = [])
 * - spatial_derivative_order:  order of discretization for 
 *                                spatial derivatives
 *                                (default = 5)
 *
 * Return value:
 * - distance_fcn:              distance function
 *
 * NOTES:
 * - All data arrays are assumed to be in the order generated by the
 *   MATLAB meshgrid() function.  That is, data corresponding to the
 *   point (x_i,y_j) is stored at index (j,i).
 *
 *===========================================================================*/

#include "mex.h"
#include "lsm_fast_marching_method.h" 

/* Input Arguments */
#define PHI	                   (prhs[0])
#define DX                         (prhs[1])
#define MASK	                   (prhs[2])
#define SPATIAL_DERIVATIVE_ORDER   (prhs[3])

/* Output Arguments */
#define DISTANCE_FUNCTION          (plhs[0])


void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  /* field data */
  double *phi;
  double *mask;
  double **ext_fields_orig;
  double *distance_fcn;
  double **ext_fields_updated;
  int num_ext_fields;
 
  /* grid data */
  const int *grid_dims = mxGetDimensions(PHI);
  double *dX = mxGetPr(DX);
  double dX_matlab_order[2];

  /* numerical parameters */
  int spatial_derivative_order;

  /* auxilliary variables */
  int i;
  int error_code;
  mxArray* tmp_mxArray;

  /* Check for proper number of arguments */
  if (nrhs < 2) {
    mexErrMsgTxt(
      "Insufficient number of input arguments (2 required; 2 optional)");
  } else if (nrhs > 4) {
    mexErrMsgTxt("Too many input arguments (2 required; 2 optional)");
  } else if (nlhs > 1) {
    mexErrMsgTxt("Too many output arguments.");
  }

  /* Get mask */
  if (nrhs < 3) {
    mask = 0;  /* NULL mask ==> all points are in interior of domain */
  } else {
    mask = mxGetPr(MASK);
  }
  
  /* Get spatial derivative order */
  if (nrhs < 4) {
    spatial_derivative_order = 1;  /* KTC - change this to 2 after */
                                   /* implementing second-order algorithm */
  } else {
    spatial_derivative_order = (int)(mxGetPr(SPATIAL_DERIVATIVE_ORDER)[0]);
  }

  /* Assign pointers for phi data */
  phi = mxGetPr(PHI);

  /* Create distance function data */
  DISTANCE_FUNCTION = mxCreateDoubleMatrix(grid_dims[0], grid_dims[1], mxREAL);
  distance_fcn = mxGetPr(DISTANCE_FUNCTION);

  /* Change order of dX to be match MATLAB meshgrid() order for grids. */
  dX_matlab_order[0] = dX[1];
  dX_matlab_order[1] = dX[0];

  /* Carry out FMM calculation */
  error_code = computeDistanceFunction2d(
                 distance_fcn,
                 phi,
                 mask,
                 spatial_derivative_order,
                 (int*) grid_dims,
                 dX_matlab_order);

  if (error_code) {
    mexErrMsgTxt("computeDistanceFunction2d failed...");
  }

  return;
}
